# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

from django.apps import AppConfig


class OtherAppConfig(AppConfig):
    name = "tests.otherapp"
    label = "tests.otherapp"
